/*
 * framecpp channel demultiplexor
 *
 * Copyright (C) 2011--2012  Kipp Cannon, Ed Maros
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


/*
 * ========================================================================
 *
 *				  Preamble
 *
 * ========================================================================
 */


/*
 * stuff from standard library
 */


#include <iostream>
#include <stdint.h>
#include <stdexcept>
#include <string.h>


/*
 * stuff from glib/gobject/gstreamer
 */


#include <glib.h>
#include <gst/gst.h>


/*
 * from framecpp
 */


#include <framecpp/Common/MemoryBuffer.hh>
#include <framecpp/Common/Verify.hh>
#include <framecpp/IFrameStream.hh>
#include <framecpp/FrameH.hh>
#include <framecpp/FrAdcData.hh>
#include <framecpp/FrProcData.hh>
#include <framecpp/FrRawData.hh>
#include <framecpp/FrSimData.hh>
#include <framecpp/FrVect.hh>


/*
 * our own stuff
 */


#include <gstlal/gstlal_debug.h>
#include <gstlal/gstlal_tags.h>
#include <framecpp_channeldemux.h>
#include <gstfrpad.h>


/*
 * ============================================================================
 *
 *                                Boilerplate
 *
 * ============================================================================
 */


#define GST_CAT_DEFAULT framecpp_channeldemux_debug
GST_DEBUG_CATEGORY_STATIC(GST_CAT_DEFAULT);


static void additional_initializations(GType type)
{
	GST_DEBUG_CATEGORY_INIT(GST_CAT_DEFAULT, "framecpp_channeldemux", 0, "framecpp_channeldemux element");
}


GST_BOILERPLATE_FULL(GstFrameCPPChannelDemux, framecpp_channeldemux, GstElement, GST_TYPE_ELEMENT, additional_initializations);


/*
 * ============================================================================
 *
 *                                 Parameters
 *
 * ============================================================================
 */


#define DEFAULT_DO_FILE_CHECKSUM FALSE
#define DEFAULT_SKIP_BAD_FILES FALSE
#define DEFAULT_FRAME_FORMAT_VERSION 0
#define DEFAULT_FRAME_LIBRARY_VERSION 255
#define DEFAULT_FRAME_LIBRARY_NAME ""
#define DEFAULT_FRAME_NAME ""
#define DEFAULT_FRAME_RUN -1
#define DEFAULT_FRAME_NUMBER 0


/*
 * ============================================================================
 *
 *                                 Utilities
 *
 * ============================================================================
 */


/* FIXME:  remove when we can rely on new-enough framecpp */
#ifndef HAVE_FRAMECPP_FrameLibraryName
static const char *get_frame_library_name(FrameCPP::IFrameStream *ifs)
{
	static const char *frame_library_names[] = {
		"unknown",
		"FrameL",
		"framecpp",
	};
	FrameCPP::Common::FrHeader::frame_library_type frame_library_type = ifs->FrameLibrary();
	g_assert_cmpint(frame_library_type, >=, 0);
	g_assert_cmpint(frame_library_type, <=, 2);
	return frame_library_names[frame_library_type];
}
#else
static const char *get_frame_library_name(FrameCPP::IFrameStream *ifs)
{
	return ifs->FrameLibraryName().c_str();
}
#endif


/*
 * split a string of the form "INSTRUMENT:CHANNEL" into two strings
 * containing the instrument and channel parts separately.  the pointers
 * placed in the instrument and channel pointers should be free()ed when no
 * longer needed.  returns 0 on success, < 0 on failure.
 */


static int split_name(const char *name, char **instrument, char **channel)
{
	const char *colon = strchr(name, ':');

	if(!colon) {
		*instrument = *channel = NULL;
		return -1;
	}

	*instrument = strndup(name, colon - name);
	*channel = strdup(colon + 1);

	return 0;
}


/*
 * check if a channel name is in the requested channel list.  returns TRUE
 * if the channel list is empty (demultiplex all channels)
 */


static gboolean is_requested_channel(GstFrameCPPChannelDemux *element, const char *name)
{
	return !g_hash_table_size(element->channel_list) || g_hash_table_lookup(element->channel_list, name);
}


/*
 * build the GstCaps describing the format of the contents of an FrVect
 */


static GstCaps *FrVect_get_caps(General::SharedPtr < FrameCPP::FrVect > vect, gint *rate, guint *unit_size, guint *bitrate)
{
	GstCaps *caps;
	gint width = vect->GetTypeSize() * 8;
	*rate = round(1.0 / vect->GetDim(0).GetDx());

	switch(vect->GetType()) {
	case FrameCPP::FrVect::FR_VECT_4R:
	case FrameCPP::FrVect::FR_VECT_8R:
		caps = gst_caps_new_simple("audio/x-raw-float",
			"rate", G_TYPE_INT, *rate,
			"channels", G_TYPE_INT, 1,
			"endianness", G_TYPE_INT, G_BYTE_ORDER,
			"width", G_TYPE_INT, width,
			NULL);
		break;

	case FrameCPP::FrVect::FR_VECT_8C:
	case FrameCPP::FrVect::FR_VECT_16C:
		caps = gst_caps_new_simple("audio/x-raw-complex",
			"rate", G_TYPE_INT, *rate,
			"channels", G_TYPE_INT, 1,
			"endianness", G_TYPE_INT, G_BYTE_ORDER,
			"width", G_TYPE_INT, width,
			NULL);
		break;

	case FrameCPP::FrVect::FR_VECT_C:
	case FrameCPP::FrVect::FR_VECT_2S:
	case FrameCPP::FrVect::FR_VECT_4S:
	case FrameCPP::FrVect::FR_VECT_8S:
		caps = gst_caps_new_simple("audio/x-raw-int",
			"rate", G_TYPE_INT, *rate,
			"channels", G_TYPE_INT, 1,
			"endianness", G_TYPE_INT, G_BYTE_ORDER,
			"width", G_TYPE_INT, width,
			"depth", G_TYPE_INT, width,	/* FIXME:  for Adc use nBits */
			"signed", G_TYPE_BOOLEAN, TRUE,
			NULL);
		break;

	case FrameCPP::FrVect::FR_VECT_1U:
	case FrameCPP::FrVect::FR_VECT_2U:
	case FrameCPP::FrVect::FR_VECT_4U:
	case FrameCPP::FrVect::FR_VECT_8U:
		caps = gst_caps_new_simple("audio/x-raw-int",
			"rate", G_TYPE_INT, *rate,
			"channels", G_TYPE_INT, 1,
			"endianness", G_TYPE_INT, G_BYTE_ORDER,
			"width", G_TYPE_INT, width,
			"depth", G_TYPE_INT, width,	/* FIXME;  for Adc use nBits */
			"signed", G_TYPE_BOOLEAN, FALSE,
			NULL);
		break;

	default:
		g_assert_not_reached();
	}

	*unit_size = 1 * width / 8;	/* 1 channel */
	*bitrate = *rate * 1 * width;	/* 1 channel */

	return caps;
}


/*
 * ============================================================================
 *
 *                                Source Pads
 *
 * ============================================================================
 */


/*
 * pad state
 */


struct pad_state {
	gboolean need_discont;
	gboolean need_new_segment;
	gboolean need_tags;
	GstClockTime next_timestamp;
	guint64 next_out_offset;
	GstTagList *tag_list;
};


/*
 * tags
 */


static GstTagList *get_pad_tag_list(GstFrPad *pad)
{
	gchar *instrument, *channel, *units;
	GstTagList *tag_list = NULL;

	g_object_get(pad, "instrument", &instrument, "channel-name", &channel, "units", &units, NULL);

	tag_list = gst_tag_list_new_full(
		GST_TAG_CODEC, "RAW",
		GST_TAG_TITLE, GST_PAD_NAME(GST_PAD_CAST(pad)),
		GSTLAL_TAG_INSTRUMENT, instrument && strcmp(instrument, "") ? instrument : " ",
		GSTLAL_TAG_CHANNEL_NAME, channel && strcmp(channel, "") ? channel : " ",
		/*GST_TAG_GEO_LOCATION_NAME, observatory,
		GST_TAG_GEO_LOCATION_SUBLOCATION, instrument,*/
		GSTLAL_TAG_UNITS, units && strcmp(units, "") ? units : " ",
		NULL
	);

	g_free(instrument);
	g_free(channel);
	g_free(units);
	return tag_list;
}


/*
 * linked event handler
 */


static void src_pad_linked(GstPad *pad, GstPad *peer, gpointer data)
{
	struct pad_state *pad_state = (struct pad_state *) gst_pad_get_element_private(pad);

	/*
	 * reset pad's state
	 */

	pad_state->need_discont = TRUE;
	pad_state->need_new_segment = TRUE;
	pad_state->need_tags = TRUE;
	pad_state->next_out_offset = 0;

	/*
	 * done
	 */
}


/*
 * create a new source pad and add to element.  does not check if name is
 * already in use
 */


static GstPad *add_pad(GstFrameCPPChannelDemux *element, const char *name, enum gst_frpad_type_t pad_type)
{
	GstFrPad *srcpad = NULL;
	struct pad_state *pad_state;
	char *instrument, *channel;

	split_name(name, &instrument, &channel);
	if(!instrument || !channel) {
		/*
		 * cannot deduce instrument and/or channel from pad's name.
		 */
		goto done;
	}

	/*
	 * construct the pad
	 */

	srcpad = gst_frpad_new_from_template(gst_element_class_get_pad_template(GST_ELEMENT_CLASS(G_OBJECT_GET_CLASS(element)), "%s"), name);
	g_assert(srcpad != NULL);
	/* FIXME:  set units properly */
	g_object_set(srcpad, "pad-type", pad_type, "instrument", instrument, "channel-name", channel, "units", strstr(channel, "STRAIN") ? "strain" : "", NULL);
	g_signal_connect(srcpad, "linked", (GCallback) src_pad_linked, NULL);

	/*
	 * create & initialize pad state.  FIXME:  this memory is
	 * leaked.  something could be attached to the pad's
	 * destroy notify signal to free the memory, but I  believe
	 * in the long run we're going to end up defining a custom
	 * subclass of GstPad for this element, and then these
	 * things can be moved into the instance structure.  it's
	 * not worth worrying about it for now.
	 */

	pad_state = g_new(struct pad_state, 1);
	g_assert(pad_state != NULL);
	pad_state->need_discont = TRUE;
	pad_state->need_new_segment = TRUE;
	pad_state->need_tags = TRUE;
	pad_state->next_timestamp = GST_CLOCK_TIME_NONE;
	pad_state->next_out_offset = 0;
	pad_state->tag_list = get_pad_tag_list(srcpad);
	g_assert(pad_state->tag_list != NULL);
	gst_pad_set_element_private(GST_PAD(srcpad), pad_state);

	/*
	 * add pad to element.  must ref it because _add_pad()
	 * consumes a reference
	 */

	gst_pad_set_active(GST_PAD(srcpad), TRUE);
	gst_object_ref(srcpad);
	gst_element_add_pad(GST_ELEMENT(element), GST_PAD(srcpad));

	/*
	 * done
	 */

done:
	g_free(instrument);
	g_free(channel);
	return GST_PAD(srcpad);
}


/*
 * remove a source pad from element
 *
 * FIXME:  what about EOS, state changes, etc.?  do we have to do that?
 */

/* NOT USED
static gboolean remove_pad(GstFrameCPPChannelDemux *element, const char *name)
{
	GstPad *srcpad = gst_element_get_static_pad(GST_ELEMENT(element), name);
	g_assert(srcpad != NULL);
	gst_pad_set_active(srcpad, FALSE);
	return gst_element_remove_pad(GST_ELEMENT(element), srcpad);
}
*/


/*
 * get pad by name, creating it if it doesn't exist
 */


static GstPad *get_src_pad(GstFrameCPPChannelDemux *element, const char *name, enum gst_frpad_type_t pad_type, gboolean *pad_added)
{
	GstPad *srcpad;

	/*
	 * retrieve the pad.  if element does not already have a pad by
	 * this name, create it
	 */

	srcpad = gst_element_get_static_pad(GST_ELEMENT(element), name);
	if(!srcpad) {
		srcpad = add_pad(element, name, pad_type);
		*pad_added = TRUE;
	} else {
		/* FIXME:  bother checking if pad_type is correct? */
	}

	return srcpad;
}


/*
 * transfer the contents of an FrVect into a newly-created GstBuffer.
 * caller must unref buffer when no longer needed.
 */


static GstBuffer *FrVect_to_GstBuffer(General::SharedPtr < FrameCPP::FrVect > vect, GstClockTime timestamp, guint64 offset, gint *rate, guint *unit_size, guint *bitrate)
{
	GstBuffer *buffer;

	/*
	 * trigger data decompression before calling GetNBytes()
	 */

	vect->GetDataUncompressed();

	/*
	 * allocate buffer
	 */

	buffer = gst_buffer_new_and_alloc(vect->GetNBytes());
	if(!buffer) {
		/* silence possibly-uninitialized warnings */
		*rate = *unit_size = *bitrate = 0;
		return NULL;
	}

	/*
	 * copy data into buffer
	 * FIXME:  it would be nice to remove the memcpy() by hooking the
	 * GstBuffer's clean-up into framecpp's reference counting
	 * machinery
	 */

	g_assert_cmpuint(vect->GetNDim(), ==, 1);
	memcpy(GST_BUFFER_DATA(buffer), vect->GetData().get(), GST_BUFFER_SIZE(buffer));

	/*
	 * set buffer format
	 */

	GST_BUFFER_CAPS(buffer) = FrVect_get_caps(vect, rate, unit_size, bitrate);
	g_assert(GST_BUFFER_CAPS(buffer) != NULL);

	/*
	 * set timestamp and duration
	 */

	GST_BUFFER_TIMESTAMP(buffer) = timestamp;
	GST_BUFFER_DURATION(buffer) = gst_util_uint64_scale_int(vect->GetNData(), GST_SECOND, *rate);
	GST_BUFFER_OFFSET(buffer) = offset;
	GST_BUFFER_OFFSET_END(buffer) = offset + vect->GetNData();

	/*
	 * done
	 */

	return buffer;
}


/*
 * convert an FrVect to a GstBuffer, and push out a source pad.
 */


static GstFlowReturn frvect_to_buffer_and_push(GstPad *pad, General::SharedPtr < FrameCPP::FrVect > vect, GstClockTime timestamp)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(gst_pad_get_parent(pad));
	struct pad_state *pad_state = (struct pad_state *) gst_pad_get_element_private(pad);
	GstBuffer *buffer;
	gint rate;
	guint unit_size;
	guint bitrate;
	GstFlowReturn result = GST_FLOW_OK;

	/*
	 * convert FrVect to GstBuffer
	 */

	buffer = FrVect_to_GstBuffer(vect, timestamp, pad_state->next_out_offset, &rate, &unit_size, &bitrate);
	g_assert(buffer != NULL);

	/*
	 * if the format matches the pad's replace the buffer's caps with
	 * the pad's to reduce the number of objects in memory and simplify
	 * subsequent comparisons.  NOTE:  the caps on the source pad get
	 * set explicitly, here, to trigger any pipeline graph adjustments
	 * that might happen as a result of the format discovery before
	 * pending segments and tags are pushed downstream.
	 */

	if(gst_caps_is_equal(GST_BUFFER_CAPS(buffer), GST_PAD_CAPS(pad)))
		gst_buffer_set_caps(buffer, GST_PAD_CAPS(pad));
	else {
		GST_LOG_OBJECT(pad, "new caps: %P", GST_BUFFER_CAPS(buffer));
		gst_pad_set_caps(pad, GST_BUFFER_CAPS(buffer));
		pad_state->need_tags = TRUE;
	}

	/*
	 * check for disconts
	 */

	if(pad_state->need_discont || (GST_CLOCK_TIME_IS_VALID(pad_state->next_timestamp) && llabs(GST_BUFFER_TIMESTAMP(buffer) - pad_state->next_timestamp) > 1)) {
		GST_BUFFER_FLAG_SET(buffer, GST_BUFFER_FLAG_DISCONT);
		pad_state->need_discont = FALSE;
	}

	/*
	 * record state for next time
	 */

	pad_state->next_timestamp = GST_BUFFER_TIMESTAMP(buffer) + GST_BUFFER_DURATION(buffer);
	pad_state->next_out_offset = GST_BUFFER_OFFSET_END(buffer);

	/*
	 * forward most recent new segment event
	 */

	if(pad_state->need_new_segment) {
		if(element->last_new_segment_event) {
			gst_event_ref(element->last_new_segment_event);
			if(!gst_pad_push_event(pad, element->last_new_segment_event))
				GST_ERROR_OBJECT(pad, "failed to push newsegment");
		}
		pad_state->need_new_segment = FALSE;
	}

	/*
	 * send tags
	 */

	if(pad_state->need_tags) {
		gst_tag_list_add(pad_state->tag_list, GST_TAG_MERGE_REPLACE, GST_TAG_BITRATE, bitrate, NULL);
		GstTagList *tag_list = gst_tag_list_merge(element->tag_list, pad_state->tag_list, GST_TAG_MERGE_REPLACE);
		GST_LOG_OBJECT(pad, "pushing %P", tag_list);
		if(!gst_pad_push_event(pad, gst_event_new_tag(tag_list)))
			GST_ERROR_OBJECT(pad, "failed to push tags");
		pad_state->need_tags = FALSE;
	}

	/*
	 * push buffer downstream
	 */

	GST_LOG_OBJECT(pad, "pushing buffer spanning %" GST_BUFFER_BOUNDARIES_FORMAT, GST_BUFFER_BOUNDARIES_ARGS(buffer));
	result = gst_pad_push(pad, buffer);

	/*
	 * done
	 */

	gst_object_unref(element);
	return result;
}


/*
 * ============================================================================
 *
 *                                  Sink Pad
 *
 * ============================================================================
 */


/*
 * event()
 */


static void gst_event_parse_new_segment_segment(GstEvent *event, GstSegment *segment)
{
	gboolean update;
	gdouble rate, applied_rate;
	GstFormat format;
	gint64 start, stop, position;

	gst_event_parse_new_segment_full(event, &update, &rate, &applied_rate, &format, &start, &stop, &position);
	gst_segment_set_newsegment_full(segment, update, rate, applied_rate, format, start, stop, position);
}


static void forward_sink_event(gpointer object, gpointer data)
{
	GstPad *pad = GST_PAD(object);
	GstEvent *event = GST_EVENT(data);
	if(gst_pad_is_linked(pad)) {
		/*
		 * forward event out this pad.  ignore failures
		 * FIXME:  should failures be ignored?
		 */
		gst_event_ref(event);
		gst_pad_push_event(pad, event);
	}
	gst_object_unref(pad);
}


static gboolean sink_event(GstPad *pad, GstEvent *event)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(gst_pad_get_parent(pad));
	GstIterator *iter;
	gboolean success = TRUE;

	switch(GST_EVENT_TYPE(event)) {
	case GST_EVENT_NEWSEGMENT:
		if(element->last_new_segment_event)
			gst_event_unref(element->last_new_segment_event);
		gst_event_ref(event);
		element->last_new_segment_event = event;
		gst_event_parse_new_segment_segment(event, &element->segment);
		break;

	case GST_EVENT_EOS:
		if(element->last_new_segment_event)
			gst_event_unref(element->last_new_segment_event);
		element->last_new_segment_event = NULL;
		break;

	default:
		break;
	}

	/* FIXME:  what does gst_pad_event_default(pad, event) do?  can I just use that? */
	iter = gst_element_iterate_src_pads(GST_ELEMENT(element));
	gst_iterator_foreach(iter, forward_sink_event, event);
	gst_iterator_free(iter);

	gst_event_unref(event);
	gst_object_unref(element);
	return success;
}


/*
 * forward_heart_beat()
 *
 * FIXME:  should check if we need to forward any pending events
 */


static GstFlowReturn push_heart_beat(GstPad *pad, GstClockTime timestamp)
{
	struct pad_state *pad_state = (struct pad_state *) gst_pad_get_element_private(pad);
	GstBuffer *buffer;

	/*
	 * create heartbeat buffer for this pad
	 */

	buffer = gst_buffer_new();
	GST_BUFFER_TIMESTAMP(buffer) = timestamp;
	GST_BUFFER_DURATION(buffer) = 0;
	GST_BUFFER_OFFSET(buffer) = GST_BUFFER_OFFSET_END(buffer) = pad_state->next_out_offset;
	gst_buffer_set_caps(buffer, GST_PAD_CAPS(pad));

	/*
	 * check for disconts
	 */

	if(pad_state->need_discont || (GST_CLOCK_TIME_IS_VALID(pad_state->next_timestamp) && llabs(GST_BUFFER_TIMESTAMP(buffer) - pad_state->next_timestamp) > 1)) {
		GST_BUFFER_FLAG_SET(buffer, GST_BUFFER_FLAG_DISCONT);
		pad_state->need_discont = FALSE;
	}

	/*
	 * record state for next time
	 */

	pad_state->next_timestamp = GST_BUFFER_TIMESTAMP(buffer) + GST_BUFFER_DURATION(buffer);
	pad_state->next_out_offset = GST_BUFFER_OFFSET_END(buffer);

	/*
	 * push buffer
	 */

	return gst_pad_push(pad, buffer);
}


static void push_heart_beat_iter_wrapper(gpointer object, gpointer data)
{
	GstPad *pad = GST_PAD(object);
	GstClockTime timestamp = *(GstClockTime *) data;

	if(gst_pad_is_linked(pad))
		/*
		 * push buffer.  ignore failures
		 * FIXME:  should failures be ignored?
		 */

		push_heart_beat(pad, timestamp);

	gst_object_unref(pad);
}


static GstFlowReturn forward_heart_beat(GstFrameCPPChannelDemux *element, GstClockTime t)
{
	GstIterator *iter = gst_element_iterate_src_pads(GST_ELEMENT(element));
	GstFlowReturn result = GST_FLOW_OK;

	gst_iterator_foreach(iter, push_heart_beat_iter_wrapper, &t);
	gst_iterator_free(iter);

	return result;
}


/*
 * chain()
 */


static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(gst_pad_get_parent(pad));
	gboolean pads_added = FALSE;
	GstPad *srcpad = NULL;
	GstFlowReturn result = GST_FLOW_OK;

	/*
	 * special case:  0-length input buffers are treated as heart
	 * beats, we forward a heardbeat out each source pad
	 */

	if(!GST_BUFFER_SIZE(inbuf)) {
		result = forward_heart_beat(element, GST_BUFFER_TIMESTAMP(inbuf));
		goto done;
	}

	/*
	 * decode frame file
	 */

	try {
		if(element->do_file_checksum) {
			/*
			 * File Checksum verification
			 *
			 * This additional scope allows for cleanup of variables used
			 * only for the file checksum validation.
			 * Resources are returned to the system as the variables go
			 * out of scope.
			 */

			FrameCPP::Common::MemoryBuffer *ibuf(new FrameCPP::Common::MemoryBuffer(std::ios::in));

			ibuf->pubsetbuf((char *) GST_BUFFER_DATA(inbuf), GST_BUFFER_SIZE(inbuf));

			FrameCPP::IFrameStream ifs(ibuf);

			FrameCPP::Common::Verify verifier;

			verifier.BufferSize(GST_BUFFER_SIZE(inbuf));
			verifier.UseMemoryMappedIO(false);
			verifier.CheckDataValid(false);	/* FIXME:  what's this? */
			verifier.Expandability(false);
			verifier.MustHaveEOFChecksum(true);
			verifier.Strict(false);
			verifier.ValidateMetadata(false);
			verifier.CheckFileChecksumOnly(true);

			if(verifier(ifs) != 0)
				throw std::runtime_error(verifier.ErrorInfo());
		}
		FrameCPP::Common::MemoryBuffer *ibuf(new FrameCPP::Common::MemoryBuffer(std::ios::in));

		ibuf->pubsetbuf((char *) GST_BUFFER_DATA(inbuf), GST_BUFFER_SIZE(inbuf));

		FrameCPP::IFrameStream ifs(ibuf);

		/*
		 * update element properties
		 */

		if(ifs.Version() != element->frame_format_version) {
			element->frame_format_version = ifs.Version();
			g_object_notify(G_OBJECT(element), "frame-format-version");
		}
		if(ifs.LibraryRevision() != element->frame_library_version) {
			element->frame_library_version = ifs.LibraryRevision();
			g_object_notify(G_OBJECT(element), "frame-library-version");
		}
		if(g_strcmp0(get_frame_library_name(&ifs), element->frame_library_name)) {
			g_free(element->frame_library_name);
			element->frame_library_name = g_strdup(get_frame_library_name(&ifs));
			g_object_notify(G_OBJECT(element), "frame-library-name");
		}

		GST_LOG_OBJECT(element, "found version %u frame file generated by %s library revision %d", element->frame_format_version, element->frame_library_name, element->frame_library_version);

		while(1) {
			FrameCPP::IFrameStream::frame_h_type frame;
			try {
				frame = ifs.ReadNextFrame();
			} catch(const std::out_of_range& Error) {
				/* no more frames */
				break;
			}

			GstClockTime frame_timestamp = 1000000000L * frame->GetGTime().GetSeconds() + frame->GetGTime().GetNanoseconds();

			GST_LOG_OBJECT(element, "found frame %d at %" GST_TIME_SECONDS_FORMAT, ifs.GetCurrentFrameOffset(), GST_TIME_SECONDS_ARGS(frame_timestamp));

			/*
			 * update element properties
			 */

			if(g_strcmp0(frame->GetName().c_str(), element->frame_name)) {
				g_free(element->frame_name);
				element->frame_name = g_strdup(frame->GetName().c_str());
				g_object_notify(G_OBJECT(element), "frame-name");
			}
			if(frame->GetRun() != element->frame_run) {
				element->frame_run = frame->GetRun();
				g_object_notify(G_OBJECT(element), "frame-run");
			}
			/* assume this changes */
			element->frame_number = frame->GetFrame();
			g_object_notify(G_OBJECT(element), "frame-number");

#if 0
			/*
			 * FIXME:  tag-related stuff that I'm not sure how to use */
			 *
			 * GPS epoch - Unix epoch = 315964800 s
			 *
			 * from pylal.xlal.datatypes import LIGOTimeGPS
			 * from pylal import date
			 * import calendar
			 * print calendar.timegm(date.XLALGPSToUTC(LIGOTimeGPS(0)))
			 */
			{
			GstDateTime *date_time = gst_date_time_new_from_unix_epoch_utc(frame_timestamp / GST_SECOND + 315964800);
			char container_format[] = "IGWD frame file v%d";	/* ok up to version 99 */
			g_assert_cmpint(ifs.Version(), <, 100);
			snprintf(container_format, sizeof(container_format), container_format, element->frame_format_version);
			gst_tag_list_add(tag_list, GST_TAG_MERGE_APPEND, GST_TAG_CONTAINER_FORMAT, container_format, GST_TAG_ENCODER, element->frame_library_name, GST_TAG_ENCODER_VERSION, element->frame_library_version, GST_TAG_ORGANIZATION, element->frame_name, NULL);
			}
#endif

			/*
			 * process ADC data
			 */

			FrameCPP::FrameH::rawData_type rd = frame->GetRawData();
			if(rd) {
				for(FrameCPP::FrRawData::firstAdc_iterator current = rd->RefFirstAdc().begin(), last = rd->RefFirstAdc().end(); current != last; current++) {
					FrameCPP::FrAdcData::data_type vects = (*current)->RefData();
					GstClockTime timestamp = frame_timestamp + (GstClockTime) round((*current)->GetTimeOffset() * 1e9);
					const char *name = (*current)->GetName().c_str();

					GST_LOG_OBJECT(element, "found FrAdcData %s at %" GST_TIME_SECONDS_FORMAT, name, GST_TIME_SECONDS_ARGS(timestamp));

					/*
					 * retrieve the source pad.  create
					 * it if it doesn't exist.  update
					 * properties to reflect stream
					 * metadata.  if the pad has no
					 * peer or it not in the requested
					 * channel list, skip this channel.
					 */

					if(!is_requested_channel(element, name)) {
						GST_LOG_OBJECT(element, "skipping: channel not requested");
						continue;
					}
					srcpad = get_src_pad(element, name, GST_FRPAD_TYPE_FRADCDATA, &pads_added);
					/* FIXME:  bias, slope, units */
					g_object_set(srcpad,
						"comment", (*current)->GetComment().c_str(),
						"channel-group", (*current)->GetChannelGroup(),
						"channel-number", (*current)->GetChannelNumber(),
						"nbits", (*current)->GetNBits(),	/* FIXME:  set depth in caps */
						NULL
					);
					if(!gst_pad_is_linked(srcpad)) {
						GST_LOG_OBJECT(element, "skipping: not linked");
						gst_object_unref(srcpad);
						srcpad = NULL;
						continue;
					}

					/*
					 * convert FrVects to GstBuffers
					 * and push out source pad,
					 * checking for disconts and
					 * recording state for next time.
					 */

					/* FIXME:  what about checking "dataValid" vect in the aux list? */
					if((*current)->GetDataValid() == 0 && vects.size()) {
						for(FrameCPP::FrAdcData::data_type::iterator vect = vects.begin(), last_vect = vects.end(); vect != last_vect; vect++) {
							result = frvect_to_buffer_and_push(srcpad, *vect, timestamp);
							if(result != GST_FLOW_OK) {
								gst_object_unref(srcpad);
								srcpad = NULL;
								GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
								goto done;
							}
						}
					} else {
						if((*current)->GetDataValid() != 0)
							GST_DEBUG_OBJECT(element, "FrAdcData invalid (dataValid=0x%04x)", (*current)->GetDataValid());
						if(!vects.size())
							GST_DEBUG_OBJECT(element, "no FrVects");
						GST_DEBUG_OBJECT(element, "pushing 0-length heart-beat buffer");
						result = push_heart_beat(srcpad, timestamp);
						if(result != GST_FLOW_OK) {
							gst_object_unref(srcpad);
							srcpad = NULL;
							GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
							goto done;
						}
					}
					gst_object_unref(srcpad);
					srcpad = NULL;
				}
			}

			/*
			 * process proc data
			 */

			for(FrameCPP::FrameH::procData_iterator current = frame->RefProcData().begin(), last = frame->RefProcData().end(); current != last; current++) {
				FrameCPP::FrProcData::data_type vects = (*current)->RefData();
				GstClockTime timestamp = frame_timestamp + (GstClockTime) round((*current)->GetTimeOffset() * 1e9);
				const char *name = (*current)->GetName().c_str();

				/*
				 * FIXME:  check the FrProcData "type"
				 * field, must be time series.  might also
				 * be able to support frequency series and
				 * time-frequency types in the future
				 */

				GST_LOG_OBJECT(element, "found FrProcData %s at %" GST_TIME_SECONDS_FORMAT, name, GST_TIME_SECONDS_ARGS(timestamp));

				/*
				 * retrieve the source pad.  create it if
				 * it doesn't exist.  update properties to
				 * reflect stream metadata.  if the pad has
				 * no peer or it not in the requested
				 * channel list, skip this channel.
				 */

				if(!is_requested_channel(element, name)) {
					GST_LOG_OBJECT(element, "skipping: channel not requested");
					continue;
				}
				srcpad = get_src_pad(element, name, GST_FRPAD_TYPE_FRPROCDATA, &pads_added);
				/* FIXME: units */
				g_object_set(srcpad,
					"comment", (*current)->GetComment().c_str(),
					NULL
				);
				if(!gst_pad_is_linked(srcpad)) {
					GST_LOG_OBJECT(element, "skipping: not linked");
					gst_object_unref(srcpad);
					srcpad = NULL;
					continue;
				}

				/*
				 * convert FrVects to GstBuffers and push
				 * out source pad, checking for disconts
				 * and recording state for next time.
				 */

				/* FIXME:  what about checking "dataValid" vect in the aux list? */
				if(vects.size()) {
					for(FrameCPP::FrProcData::data_type::iterator vect = vects.begin(), last_vect = vects.end(); vect != last_vect; vect++) {
						result = frvect_to_buffer_and_push(srcpad, *vect, timestamp);
						if(result != GST_FLOW_OK) {
							gst_object_unref(srcpad);
							srcpad = NULL;
							GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
							goto done;
						}
					}
				} else {
					GST_DEBUG_OBJECT(element, "no FrVects.  pushing 0-length heart-beat buffer");
					result = push_heart_beat(srcpad, timestamp);
					if(result != GST_FLOW_OK) {
						gst_object_unref(srcpad);
						srcpad = NULL;
						GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
						goto done;
					}
				}
				gst_object_unref(srcpad);
				srcpad = NULL;
			}

			/*
			 * process simulated data
			 */

			for(FrameCPP::FrameH::simData_iterator current = frame->RefSimData().begin(), last = frame->RefSimData().end(); current != last; current++) {
				FrameCPP::FrSimData::data_type vects = (*current)->RefData();
				GstClockTime timestamp = frame_timestamp + (GstClockTime) round((*current)->GetTimeOffset() * 1e9);
				const char *name = (*current)->GetName().c_str();

				GST_LOG_OBJECT(element, "found FrSimData %s at %" GST_TIME_SECONDS_FORMAT, name, GST_TIME_SECONDS_ARGS(timestamp));

				/*
				 * retrieve the source pad.  create it if
				 * it doesn't exist.  update properties to
				 * reflect stream metadata.  if the pad has
				 * no peer or it not in the requested
				 * channel list, skip this channel.
				 */

				if(!is_requested_channel(element, name)) {
					GST_LOG_OBJECT(element, "skipping: channel not requested");
					continue;
				}
				srcpad = get_src_pad(element, name, GST_FRPAD_TYPE_FRSIMDATA, &pads_added);
				/* FIXME: units */
				g_object_set(srcpad,
					"comment", (*current)->GetComment().c_str(),
					NULL
				);
				if(!gst_pad_is_linked(srcpad)) {
					GST_LOG_OBJECT(element, "skipping: not linked");
					gst_object_unref(srcpad);
					srcpad = NULL;
					continue;
				}

				/*
				 * convert FrVect to a GstBuffer and push
				 * out source pad, checking for disconts
				 * and recording state for next time.
				 */

				/* FIXME:  what about checking "dataValid" vect in the aux list? */
				if(vects.size()) {
					for(FrameCPP::FrSimData::data_type::iterator vect = vects.begin(), last_vect = vects.end(); vect != last_vect; vect++) {
						result = frvect_to_buffer_and_push(srcpad, *vect, timestamp);
						if(result != GST_FLOW_OK) {
							gst_object_unref(srcpad);
							srcpad = NULL;
							GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
							goto done;
						}
					}
				} else {
					GST_DEBUG_OBJECT(element, "no FrVects.  pushing 0-length heart-beat buffer");
					result = push_heart_beat(srcpad, timestamp);
					if(result != GST_FLOW_OK) {
						gst_object_unref(srcpad);
						srcpad = NULL;
						GST_ERROR_OBJECT(element, "failure: %s", gst_flow_get_name(result));
						goto done;
					}
				}
				gst_object_unref(srcpad);
				srcpad = NULL;
			}
		}
	} catch(const std::exception& Exception) {
		if(srcpad)
			gst_object_unref(srcpad);
		if(element->skip_bad_files)
			GST_ELEMENT_WARNING(element, STREAM, DECODE, (NULL), ("libframecpp raised exception: %s", Exception.what()));
		else {
			GST_ELEMENT_ERROR(element, STREAM, DECODE, (NULL), ("libframecpp raised exception: %s", Exception.what()));
			result = GST_FLOW_ERROR;
			goto done;
		}
	} catch(...) {
		if(srcpad)
			gst_object_unref(srcpad);
		if(element->skip_bad_files)
			GST_ELEMENT_WARNING(element, STREAM, DECODE, (NULL), ("libframecpp raised unknown exception"));
		else {
			GST_ELEMENT_ERROR(element, STREAM, DECODE, (NULL), ("libframecpp raised unknown exception"));
			result = GST_FLOW_ERROR;
			goto done;
		}
	}

	/*
	 * FIXME:  should src pads whose names weren't found in this frame
	 * file be removed?  this can also handle the case of the requested
	 * channel list having names removed from it
	 */

	/*
	 * Done
	 */

done:
	if(pads_added)
		gst_element_no_more_pads(GST_ELEMENT(element));
	gst_buffer_unref(inbuf);
	gst_object_unref(element);
	return result;
}


/*
 * ============================================================================
 *
 *                              GObject Methods
 *
 * ============================================================================
 */


/*
 * properites
 */


enum property {
	ARG_DO_FILE_CHECKSUM = 1,
	ARG_SKIP_BAD_FILES,
	ARG_CHANNEL_LIST,
	ARG_FRAME_FORMAT_VERSION,
	ARG_FRAME_LIBRARY_VERSION,
	ARG_FRAME_LIBRARY_NAME,
	ARG_FRAME_NAME,
	ARG_FRAME_RUN,
	ARG_FRAME_NUMBER
};


static void set_property(GObject *object, guint id, const GValue *value, GParamSpec *pspec)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(object);

	GST_OBJECT_LOCK(element);

	switch((enum property) id) {
	case ARG_DO_FILE_CHECKSUM:
		element->do_file_checksum = g_value_get_boolean(value);
		break;

	case ARG_SKIP_BAD_FILES:
		element->skip_bad_files = g_value_get_boolean(value);
		break;

	case ARG_CHANNEL_LIST: {
		GValueArray *channel_list = (GValueArray *) g_value_get_boxed(value);
		guint i;

		g_hash_table_remove_all(element->channel_list);

		for(i = 0; i < channel_list->n_values; i++) {
			gchar *channel_name = g_value_dup_string(g_value_array_get_nth(channel_list, i));
			g_hash_table_replace(element->channel_list, channel_name, channel_name);
		}

		break;
	}

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, id, pspec);
		break;
	}

	GST_OBJECT_UNLOCK(element);
}


static void get_property(GObject *object, guint id, GValue *value, GParamSpec *pspec)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(object);

	GST_OBJECT_LOCK(element);

	switch((enum property) id) {
	case ARG_DO_FILE_CHECKSUM:
		g_value_set_boolean(value, element->do_file_checksum);
		break;

	case ARG_SKIP_BAD_FILES:
		g_value_set_boolean(value, element->skip_bad_files);
		break;

	case ARG_CHANNEL_LIST: {
		GValueArray *channel_list = g_value_array_new(0);
		GValue channel_name = {0};
		GHashTableIter iter;
		gchar *key, *ignored;

		g_value_init(&channel_name, G_TYPE_STRING);
		g_hash_table_iter_init(&iter, element->channel_list);

		while(g_hash_table_iter_next(&iter, (void **) &key, (void **) &ignored)) {
			g_value_set_string(&channel_name, key);
			g_value_array_append(channel_list, &channel_name);
			g_value_reset(&channel_name);
		}

		g_value_take_boxed(value, channel_list);
		break;
	}

	case ARG_FRAME_FORMAT_VERSION:
		g_value_set_uint(value, element->frame_format_version);
		break;

	case ARG_FRAME_LIBRARY_VERSION:
		g_value_set_uint(value, element->frame_library_version);
		break;

	case ARG_FRAME_LIBRARY_NAME:
		g_value_set_string(value, element->frame_library_name);
		break;

	case ARG_FRAME_NAME:
		g_value_set_string(value, element->frame_name);
		break;

	case ARG_FRAME_RUN:
		g_value_set_int(value, element->frame_run);
		break;

	case ARG_FRAME_NUMBER:
		g_value_set_uint(value, element->frame_number);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, id, pspec);
		break;
	}

	GST_OBJECT_UNLOCK(element);
}


/*
 * finalize()
 */


static void finalize(GObject * object)
{
	GstFrameCPPChannelDemux *element = FRAMECPP_CHANNELDEMUX(object);

	if(element->last_new_segment_event)
		gst_event_unref(element->last_new_segment_event);
	element->last_new_segment_event = NULL;
	g_hash_table_unref(element->channel_list);
	element->channel_list = NULL;
	gst_tag_list_free(element->tag_list);
	element->tag_list = NULL;
	g_free(element->frame_library_name);
	element->frame_library_name = NULL;
	g_free(element->frame_name);
	element->frame_name = NULL;

	G_OBJECT_CLASS(parent_class)->finalize(object);
}


/*
 * Pad templates.
 */


static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
	"%s",
	GST_PAD_SRC,
	GST_PAD_SOMETIMES,
	GST_STATIC_CAPS(
		"audio/x-raw-float, " \
			"rate = (int) [1, MAX], " \
			"channels = (int) 1, " \
			"endianness = (int) BYTE_ORDER, " \
			"width = (int) {32, 64}; "\
		"audio/x-raw-int, " \
			"rate = (int) [1, MAX], " \
			"channels = (int) 1, " \
			"endianness = (int) BYTE_ORDER, " \
			"width = (int) 8, " \
			"depth = (int) 8, " \
			"signed = (boolean) {true, false};" \
		"audio/x-raw-int, " \
			"rate = (int) [1, MAX], " \
			"channels = (int) 1, " \
			"endianness = (int) BYTE_ORDER, " \
			"width = (int) 16, " \
			"depth = (int) 16, " \
			"signed = (boolean) {true, false};" \
		"audio/x-raw-int, " \
			"rate = (int) [1, MAX], " \
			"channels = (int) 1, " \
			"endianness = (int) BYTE_ORDER, " \
			"width = (int) 32, " \
			"depth = (int) 32, " \
			"signed = (boolean) {true, false};" \
		"audio/x-raw-int, " \
			"rate = (int) [1, MAX], " \
			"channels = (int) 1, " \
			"endianness = (int) BYTE_ORDER, " \
			"width = (int) 64, " \
			"depth = (int) 64, " \
			"signed = (boolean) {true, false};" \
	)
);


/*
 * base_init()
 */


static void framecpp_channeldemux_base_init(gpointer klass)
{
}


/*
 * class_init()
 */


static void framecpp_channeldemux_class_init(GstFrameCPPChannelDemuxClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS(klass);

	gobject_class->set_property = GST_DEBUG_FUNCPTR(set_property);
	gobject_class->get_property = GST_DEBUG_FUNCPTR(get_property);
	gobject_class->finalize = GST_DEBUG_FUNCPTR(finalize);

	gst_element_class_set_details_simple(
		element_class,
		"IGWD frame file channel demuxer",
		"Codec/Demuxer",
		"demux streams from IGWD frame files (https://dcc.ligo.org/cgi-bin/DocDB/ShowDocument?docid=329)",
		"Kipp Cannon <kipp.cannon@ligo.org>, Ed Maros <ed.maros@ligo.org>"
	);

	gst_element_class_add_pad_template(
		element_class,
		gst_pad_template_new(
			"sink",
			GST_PAD_SINK,
			GST_PAD_ALWAYS,
			gst_caps_new_simple(
				"application/x-igwd-frame",
				"framed", G_TYPE_BOOLEAN, TRUE,
				NULL
			)
		)
	);

	gst_element_class_add_pad_template(
		element_class,
		gst_static_pad_template_get(&src_factory)
	);

	g_object_class_install_property(
		gobject_class,
		ARG_DO_FILE_CHECKSUM,
		g_param_spec_boolean(
			"do-file-checksum",
			"Do file checksum",
			"Checks the file-level checksum of each input file (individual structure checksums are always checked).  This is costly for large (e.g., level 0) frame files, so it is disabled by default.",
			DEFAULT_DO_FILE_CHECKSUM,
			(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_SKIP_BAD_FILES,
		g_param_spec_boolean(
			"skip-bad-files",
			"Ignore bad files",
			"Treat files that fail validation checks as though they are missing instead of raising an error.  The next buffers to be demultiplexed will be marked as discontinuities.",
			DEFAULT_SKIP_BAD_FILES,
			(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_CHANNEL_LIST,
		g_param_spec_value_array(
			"channel-list",
			"Channel list",
			"Restrict demultiplexed channels to those in this list.  An empty list (default) causes all channels to be demultiplexed.  The use of this feature can improve performance when demultiplexing files with large numbers of channels;  it can be ignored for small files.  It is not an error for names in this list to not appear in the frame files.",
			g_param_spec_string(
				"channel",
				"Channel name",
				"Name of channel to demultiplex.",
				NULL,
				(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
			),
			(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_FORMAT_VERSION,
		g_param_spec_uint(
			"frame-format-version",
			"Frame format version",
			"Format version number from file header.",
			0, 255, DEFAULT_FRAME_FORMAT_VERSION,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_LIBRARY_VERSION,
		g_param_spec_uint(
			"frame-library-version",
			"Frame library version",
			"Frame library version from file header.",
			0, 255, DEFAULT_FRAME_LIBRARY_VERSION,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_LIBRARY_NAME,
		g_param_spec_string(
			"frame-library-name",
			"Frame library name",
			"Frame library name from file header.",
			DEFAULT_FRAME_LIBRARY_NAME,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_NAME,
		g_param_spec_string(
			"frame-name",
			"Frame name",
			"Name appearing in frame header.",
			DEFAULT_FRAME_NAME,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
	G_PARAM_SPEC_STRING(g_object_class_find_property(gobject_class, "frame-name"))->ensure_non_null = TRUE;
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_RUN,
		g_param_spec_int(
			"frame-run",
			"Run number",
			"Run number appearing in frame header.",
			G_MININT, G_MAXINT, DEFAULT_FRAME_RUN,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FRAME_NUMBER,
		g_param_spec_uint(
			"frame-number",
			"Frame number",
			"Current frame number.",
			0, G_MAXUINT, DEFAULT_FRAME_NUMBER,
			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
		)
	);
}


/*
 * instance_init()
 */


static void framecpp_channeldemux_init(GstFrameCPPChannelDemux *element, GstFrameCPPChannelDemuxClass *klass)
{
	GstPad *pad;

	gst_element_create_all_pads(GST_ELEMENT(element));

	/* configure sink pad */
	pad = gst_element_get_static_pad(GST_ELEMENT(element), "sink");
	gst_pad_set_chain_function(pad, GST_DEBUG_FUNCPTR(chain));
	gst_pad_set_event_function(pad, GST_DEBUG_FUNCPTR(sink_event));
	gst_object_unref(pad);

	/* internal data */
	element->last_new_segment_event = NULL;
	element->channel_list = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
	element->tag_list = gst_tag_list_new();
	element->frame_format_version = DEFAULT_FRAME_FORMAT_VERSION;
	element->frame_library_version = DEFAULT_FRAME_LIBRARY_VERSION;
	element->frame_library_name = g_strdup(DEFAULT_FRAME_LIBRARY_NAME);
	element->frame_name = g_strdup(DEFAULT_FRAME_NAME);
	element->frame_run = DEFAULT_FRAME_RUN;
	element->frame_number = DEFAULT_FRAME_NUMBER;
}
